"""MediaFireUploader - API encapsulating Upload magic"""

from __future__ import unicode_literals

import hashlib
import logging
import math
import os
import time

from collections import namedtuple

from mediafire.subsetio import SubsetIO
from mediafire.api import MediaFireConnectionError

MEBIBYTE = 2 ** 20

# Use resumable upload if file is larger than 4MiB
UPLOAD_SIMPLE_LIMIT_BYTES = 4 * MEBIBYTE

# Retry resumable uploads 5 times
UPLOAD_RETRY_COUNT = 5

# Upload polling interval in seconds
UPLOAD_POLL_INTERVAL = 5

# Length of upload key
UPLOAD_KEY_LENGTH = 11

# File upload statuses
STATUS_NO_MORE_REQUESTS = 99
STATUS_UPLOAD_IN_PROGRESS = 17

# Read this much during hashing, must be a power of 2 and not more than 2 ** 10
HASH_CHUNK_SIZE_BYTES = 8192

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


# pylint: disable=too-few-public-methods,too-many-arguments
# pylint: disable=too-many-instance-attributes
class _UploadInfo(object):
    """Structure containing upload details"""

    def __init__(self, fd=None, name=None, folder_key=None, path=None,
                 hash_info=None, size=None, filedrop_key=None,
                 action_on_duplicate=None):
        self.fd = fd
        self.name = name
        self.folder_key = folder_key
        self.path = path
        self.hash_info = hash_info
        self.size = size
        self.filedrop_key = filedrop_key
        self.action_on_duplicate = action_on_duplicate


class _UploadUnitInfo(object):
    """Structure containing upload unit details"""
    def __init__(self, upload_info=None, fd=None, uid=None, hash_=None):
        self.upload_info = upload_info
        self.fd = fd
        self.uid = uid
        self.hash_ = hash_


UploadResult = namedtuple('UploadResult', [
    'action', 'quickkey', 'hash_', 'filename', 'size', 'created', 'revision'
])


# pylint: enable=too-few-public-methods,too-many-arguments
class UploadSession(object):  # pylint: disable=too-few-public-methods
    """Allocate/deallocate action token automatically"""

    def __init__(self, api):
        """Initialize context manager

        api -- MediaFireApi instance
        """
        self.action_token = None
        self._api = api

    def __enter__(self):
        """Allocate action token"""
        self.action_token = self._api.user_get_action_token(
            type_="upload", lifespan=1440)['action_token']

        self._api.set_action_token(type_="upload",
                                   action_token=self.action_token)

    def __exit__(self, *exc_details):
        """Destroys action token"""
        self._api.user_destroy_action_token(action_token=self.action_token)


class UploadError(Exception):
    """Basic upload error"""
    pass


class RetriableUploadError(UploadError):
    """Retriable upload error"""
    pass


MediaFireHashInfo = namedtuple('MediaFireHashInfo', [
    # sha256 digest of the whole file
    'file',
    # array of sha256 digests of upload units
    'units',
    # size of the file, for verification
    'size'
])


def decode_resumable_upload_bitmap(bitmap_node, number_of_units):
    """Decodes bitmap_node to hash of unit_id: is_uploaded

    bitmap_node -- bitmap node of resumable_upload with
                   'count' number and 'words' containing array
    number_of_units -- number of units we are uploading to
                       define the number of bits for bitmap
    """
    bitmap = 0
    for token_id in range(int(bitmap_node['count'])):
        value = int(bitmap_node['words'][token_id])
        bitmap = bitmap | (value << (0xf * token_id))

    result = {}

    for unit_id in range(number_of_units):
        mask = 1 << unit_id
        result[unit_id] = (bitmap & mask) == mask

    return result


def compute_resumable_upload_unit_size(size):
    """Calculate resumable upload unit size from file size

    size -- size of the file in bytes
    """

    size_in_mib = math.floor(size / MEBIBYTE)

    if size_in_mib > 1:
        # stepping = log_4MiB(size_MiB)
        threshold_stepping = int(
            math.floor(math.log(size_in_mib, 4))
        )
        # 16 GB (7) is the last threshold stepping
        if threshold_stepping > 7:
            threshold_stepping = 7
    else:
        threshold_stepping = 1

    # Ux = 1024 * 2^(X-1)
    # Ux is in kilobytes, so we multiply by 1024 to get bytes
    unit_size = 1024 * 1024 * 2 ** (threshold_stepping - 1)
    return unit_size


def compute_hash_info(fd, unit_size=None):
    """Get MediaFireHashInfo structure from the fd, unit_size

    fd -- file descriptor - expects exclusive access because of seeking
    unit_size -- size of a single unit

    Returns MediaFireHashInfo:
    hi.file -- sha256 of the whole file
    hi.units -- list of sha256 hashes for each unit
    """

    fd.seek(0, os.SEEK_END)
    file_size = fd.tell()
    fd.seek(0, os.SEEK_SET)

    units = []
    unit_counter = 0

    file_hash = hashlib.sha256()
    unit_hash = hashlib.sha256()

    for chunk in iter(lambda: fd.read(HASH_CHUNK_SIZE_BYTES), b''):
        file_hash.update(chunk)

        unit_hash.update(chunk)
        unit_counter += len(chunk)

        if unit_size is not None and unit_counter == unit_size:
            # flush the current unit hash
            units.append(unit_hash.hexdigest().lower())
            unit_counter = 0
            unit_hash = hashlib.sha256()

    if unit_size is not None and unit_counter > 0:
        # leftover block
        units.append(unit_hash.hexdigest().lower())

    fd.seek(0, os.SEEK_SET)

    return MediaFireHashInfo(
        file=file_hash.hexdigest().lower(),
        units=units,
        size=file_size
    )


class MediaFireUploader(object):
    """API encapsulating Upload magic"""

    def __init__(self, api):
        """Initialize MediaFireUploader

        api -- MediaFireApi instance
        """
        self._api = api

    # pylint: disable=too-many-arguments
    def upload(self, fd, name=None, folder_key=None, filedrop_key=None,
               path=None, hash_info=None, action_on_duplicate=None):
        """Upload file, returns UploadResult object

        fd -- file-like object to upload from, expects exclusive access
        name -- file name
        folder_key -- folderkey of the target folder
        path -- path to file relative to folder_key
        hash_info -- MediaFireHashInfo for file contents
        filedrop_key -- filedrop to use instead of folder_key
        action_on_duplicate -- skip, keep, replace
        """

        # Get file handle content length in the most reliable way
        fd.seek(0, os.SEEK_END)
        size = fd.tell()
        fd.seek(0, os.SEEK_SET)

        if size > UPLOAD_SIMPLE_LIMIT_BYTES:
            resumable = True
            unit_size = compute_resumable_upload_unit_size(size)
        else:
            resumable = False
            unit_size = None

        # Allow supplying stored HashInfo
        if hash_info is None:
            logger.debug("Calculating checksum")
            hash_info = compute_hash_info(fd, unit_size)

        if hash_info.size != size:
            # Has the file changed beween computing the hash
            # and calling upload()?
            raise ValueError("hash_info.size mismatch")

        upload_info = _UploadInfo(fd=fd, name=name, folder_key=folder_key,
                                  hash_info=hash_info, size=size, path=path,
                                  filedrop_key=filedrop_key,
                                  action_on_duplicate=action_on_duplicate)

        # Check whether file is present
        check_result = self._upload_check(upload_info, resumable)

        upload_result = None
        upload_func = None

        folder_key = check_result.get('folder_key', None)
        if folder_key is not None:
            # We know precisely what folder_key to use, drop path
            upload_info.folder_key = folder_key
            upload_info.path = None

        if check_result['hash_exists'] == 'yes':
            # file exists somewhere in MediaFire
            if check_result['in_folder'] == 'yes' and \
                    check_result['file_exists'] == 'yes':
                # file exists in this directory
                different_hash = check_result.get('different_hash', 'no')
                if different_hash == 'no':
                    # file is already there
                    upload_func = self._upload_none

            if not upload_func:
                # different hash or in other folder
                upload_func = self._upload_instant

        if not upload_func:
            if resumable:
                upload_func = self._upload_resumable
            else:
                upload_func = self._upload_simple

        # Retry retriable exceptions
        retries = UPLOAD_RETRY_COUNT
        while retries > 0:
            try:
                # Provide check_result to avoid calling API twice
                upload_result = upload_func(upload_info, check_result)
            except (RetriableUploadError, MediaFireConnectionError):
                retries -= 1
                logger.exception("%s failed (%d retries left)",
                                 upload_func.__name__, retries)
                # Refresh check_result for next iteration
                check_result = self._upload_check(upload_info, resumable)
            except Exception:
                logger.exception("%s failed", upload_func)
                break
            else:
                break

        if upload_result is None:
            raise UploadError("Upload failed")

        return upload_result
    # pylint: enable=too-many-arguments

    def _poll_upload(self, upload_key, action):
        """Poll upload until quickkey is found

        upload_key -- upload_key returned by upload/* functions
        """

        if len(upload_key) != UPLOAD_KEY_LENGTH:
            # not a regular 11-char-long upload key
            # There is no API to poll filedrop uploads
            return UploadResult(
                action=action,
                quickkey=None,
                hash_=None,
                filename=None,
                size=None,
                created=None,
                revision=None
            )

        quick_key = None
        while quick_key is None:
            poll_result = self._api.upload_poll(upload_key)
            doupload = poll_result['doupload']

            logger.debug("poll(%s): status=%d, description=%s, filename=%s,"
                         " result=%d",
                         upload_key, int(doupload['status']),
                         doupload['description'], doupload['filename'],
                         int(doupload['result']))

            if int(doupload['result']) != 0:
                break

            if doupload['fileerror'] != '':
                # TODO: we may have to handle this a bit more dramatically
                logger.warning("poll(%s): fileerror=%d", upload_key,
                               int(doupload['fileerror']))
                break

            if int(doupload['status']) == STATUS_NO_MORE_REQUESTS:
                quick_key = doupload['quickkey']
            elif int(doupload['status']) == STATUS_UPLOAD_IN_PROGRESS:
                # BUG: http://forum.mediafiredev.com/showthread.php?588
                raise RetriableUploadError(
                    "Invalid state transition ({})".format(
                        doupload['description']
                    )
                )
            else:
                time.sleep(UPLOAD_POLL_INTERVAL)

        return UploadResult(
            action=action,
            quickkey=doupload['quickkey'],
            hash_=doupload['hash'],
            filename=doupload['filename'],
            size=doupload['size'],
            created=doupload['created'],
            revision=doupload['revision']
        )

    def _upload_check(self, upload_info, resumable=False):
        """Wrapper around upload/check"""
        return self._api.upload_check(
            filename=upload_info.name,
            size=upload_info.size,
            hash_=upload_info.hash_info.file,
            folder_key=upload_info.folder_key,
            filedrop_key=upload_info.filedrop_key,
            path=upload_info.path,
            resumable=resumable
        )

    # pylint: disable=no-self-use
    # We just provide a consistent interface
    def _upload_none(self, upload_info, check_result):
        """Dummy upload function for when we don't actually upload"""
        return UploadResult(
            action=None,
            quickkey=check_result['duplicate_quickkey'],
            hash_=upload_info.hash_info.file,
            filename=upload_info.name,
            size=upload_info.size,
            created=None,
            revision=None
        )
    # pylint: enable=no-self-use

    def _upload_instant(self, upload_info, _=None):
        """Instant upload and return quickkey

        Can be used when the file is already stored somewhere in MediaFire

        upload_info -- UploadInfo object
        check_result -- ignored
        """

        result = self._api.upload_instant(
            upload_info.name,
            upload_info.size,
            upload_info.hash_info.file,
            path=upload_info.path,
            folder_key=upload_info.folder_key,
            filedrop_key=upload_info.filedrop_key,
            action_on_duplicate=upload_info.action_on_duplicate
        )

        return UploadResult(
            action='upload/instant',
            quickkey=result['quickkey'],
            filename=result['filename'],
            revision=result['new_device_revision'],
            hash_=upload_info.hash_info.file,
            size=upload_info.size,
            created=None
        )

    def _upload_simple(self, upload_info, _=None):
        """Simple upload and return quickkey

        Can be used for small files smaller than UPLOAD_SIMPLE_LIMIT_BYTES

        upload_info -- UploadInfo object
        check_result -- ignored
        """

        upload_result = self._api.upload_simple(
            upload_info.fd,
            upload_info.name,
            folder_key=upload_info.folder_key,
            filedrop_key=upload_info.filedrop_key,
            path=upload_info.path,
            file_size=upload_info.size,
            file_hash=upload_info.hash_info.file,
            action_on_duplicate=upload_info.action_on_duplicate)

        logger.debug("upload_result: %s", upload_result)

        upload_key = upload_result['doupload']['key']

        return self._poll_upload(upload_key, 'upload/simple')

    def _upload_resumable_unit(self, uu_info):
        """Upload a single unit and return raw upload/resumable result

        uu_info -- UploadUnitInfo instance
        """

        # Get actual unit size
        unit_size = uu_info.fd.len

        if uu_info.hash_ is None:
            raise ValueError('UploadUnitInfo.hash_ is now required')

        return self._api.upload_resumable(
            uu_info.fd,
            uu_info.upload_info.size,
            uu_info.upload_info.hash_info.file,
            uu_info.hash_,
            uu_info.uid,
            unit_size,
            filedrop_key=uu_info.upload_info.filedrop_key,
            folder_key=uu_info.upload_info.folder_key,
            path=uu_info.upload_info.path,
            action_on_duplicate=uu_info.upload_info.action_on_duplicate)

    def _upload_resumable_all(self, upload_info, bitmap,
                              number_of_units, unit_size):
        """Prepare and upload all resumable units and return upload_key

        upload_info -- UploadInfo object
        bitmap -- bitmap node of upload/check
        number_of_units -- number of units requested
        unit_size -- size of a single upload unit in bytes
        """

        fd = upload_info.fd

        upload_key = None

        for unit_id in range(number_of_units):
            upload_status = decode_resumable_upload_bitmap(
                bitmap, number_of_units)

            if upload_status[unit_id]:
                logger.debug("Skipping unit %d/%d - already uploaded",
                             unit_id + 1, number_of_units)
                continue

            logger.debug("Uploading unit %d/%d",
                         unit_id + 1, number_of_units)

            offset = unit_id * unit_size

            with SubsetIO(fd, offset, unit_size) as unit_fd:

                unit_info = _UploadUnitInfo(
                    upload_info=upload_info,
                    hash_=upload_info.hash_info.units[unit_id],
                    fd=unit_fd,
                    uid=unit_id)

                upload_result = self._upload_resumable_unit(unit_info)

                # upload_key is needed for polling
                if upload_key is None:
                    upload_key = upload_result['doupload']['key']

        return upload_key

    def _upload_resumable(self, upload_info, check_result):
        """Resumable upload and return quickkey

        upload_info -- UploadInfo object
        check_result -- dict of upload/check call result
        """

        resumable_upload = check_result['resumable_upload']

        unit_size = int(resumable_upload['unit_size'])
        number_of_units = int(resumable_upload['number_of_units'])

        # make sure we have calculated the right thing
        assert len(upload_info.hash_info.units) == number_of_units
        assert(unit_size ==
               compute_resumable_upload_unit_size(upload_info.size))

        logger.debug("Preparing %d units * %d bytes",
                     number_of_units, unit_size)

        upload_key = None
        retries = UPLOAD_RETRY_COUNT

        all_units_ready = resumable_upload['all_units_ready'] == 'yes'
        bitmap = resumable_upload['bitmap']

        while not all_units_ready and retries > 0:
            upload_key = self._upload_resumable_all(upload_info, bitmap,
                                                    number_of_units, unit_size)

            check_result = self._upload_check(upload_info, resumable=True)

            resumable_upload = check_result['resumable_upload']
            all_units_ready = resumable_upload['all_units_ready'] == 'yes'
            bitmap = resumable_upload['bitmap']

            if not all_units_ready:
                retries -= 1
                logger.debug("Some units failed to upload (%d retries left)",
                             retries)

        if not all_units_ready:
            # Most likely non-retriable
            raise UploadError("Could not upload all units")

        logger.debug("Upload complete, polling for status")

        return self._poll_upload(upload_key, 'upload/resumable')



"""SubsetIO"""

from __future__ import unicode_literals

import os
import io
import logging


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class SubsetIO(io.IOBase):
    """minimal file-like object exposing subset of parent file"""
    def __init__(self, fd, offset, length):
        """Create new ChunkedFileWrapper object

        fd -- parent file handle
        offset -- start byte of the view
        length -- length of the view
        """

        try:
            self.parent_fd = os.fdopen(os.dup(fd.fileno()), 'rb')
        except io.UnsupportedOperation:
            logger.debug("Re-using parent fd (not thread-safe)")
            self.parent_fd = fd

        self.offset = offset
        # name also makes requests.utils.super_len() work
        self.len = length

        # find the size of the original file
        self.parent_fd.seek(0, os.SEEK_END)
        file_size = self.parent_fd.tell()

        if self.offset + self.len > file_size:
            self.len = file_size - self.offset

        if self.offset < 0:
            self.offset = 0

        if self.len < 0:
            self.len = 0

        self.parent_fd.seek(self.offset)

    def read(self, limit=-1):
        """Read content. See file.read"""
        remaining = self.len - self.parent_fd.tell() + self.offset

        if limit > remaining or limit == -1:
            limit = remaining

        return self.parent_fd.read(limit)

    def seek(self, offset, whence=os.SEEK_SET):
        """Seek to position in stream, see file.seek"""

        pos = None
        if whence == os.SEEK_SET:
            pos = self.offset + offset
        elif whence == os.SEEK_CUR:
            pos = self.tell() + offset
        elif whence == os.SEEK_END:
            pos = self.offset + self.len + offset
        else:
            raise ValueError("invalid whence {}".format(whence))

        if pos > self.offset + self.len or pos < self.offset:
            raise ValueError("seek position beyond chunk area")

        self.parent_fd.seek(pos, os.SEEK_SET)

    def tell(self):
        """Get current position in file, see file.tell"""
        result = self.parent_fd.tell() - self.offset
        return result

    def close(self):
        """Close file, see file.close"""
        try:
            self.parent_fd.fileno()
        except io.UnsupportedOperation:
            logger.debug("Not closing parent_fd - reusing existing")
        else:
            self.parent_fd.close()



"""Conversion Server API"""
# Audio and Video transcoding URL is sent in response to file/get_links
# and only Image and Document conversion use conversion_server.php endpoint

from __future__ import unicode_literals

import logging
import requests

from six.moves.urllib.parse import urlencode

from mediafire.api import QueryParams

logger = logging.getLogger(__name__)

API_ENDPOINT = 'https://www.mediafire.com/conversion_server.php'


class ConversionServerError(Exception):
    """Conversion Server Errors"""
    def __init__(self, message, status):
        self.message = message
        self.status = status
        super(ConversionServerError, self).__init__(message, status)

    def __str__(self):
        return "{}: {}".format(self.status, self.message)


class ConversionServerClient(object):
    """Conversion Server client"""

    def __init__(self):
        self.http = requests.Session()

    def request(self, hash_, quickkey, doc_type, page=None,
                output=None, size_id=None, metadata=None,
                request_conversion_only=None):
        """Query conversion server

        hash_: 4 characters of file hash
        quickkey: File quickkey
        doc_type: "i" for image, "d" for documents
        page: The page to convert. If page is set to 'initial', the first
              10 pages of the document will be provided. (document)
        output: "pdf", "img", or "swf" (document)
        size_id: 0,1,2 (document)
                 0-9, a-f, z (image)
        metadata: Set to 1 to get metadata dict
        request_conversion_only: Request conversion w/o content
        """

        if len(hash_) > 4:
            hash_ = hash_[:4]

        query = QueryParams({
            'quickkey': quickkey,
            'doc_type': doc_type,
            'page': page,
            'output': output,
            'size_id': size_id,
            'metadata': metadata,
            'request_conversion_only': request_conversion_only
        })

        url = API_ENDPOINT + '?' + hash_ + '&' + urlencode(query)

        response = self.http.get(url, stream=True)

        if response.status_code == 204:
            raise ConversionServerError("Unable to fulfill request. "
                                        "The document will not be converted.",
                                        response.status_code)

        response.raise_for_status()

        if response.headers['content-type'] == 'application/json':
            return response.json()

        return response



"""Low-level MediaFire API Client"""

from __future__ import unicode_literals

import hashlib
import requests
import logging

import six

from six.moves.urllib.parse import urlencode

from requests_toolbelt import MultipartEncoder
from requests.adapters import HTTPAdapter

from requests.exceptions import RequestException

API_BASE = 'https://www.mediafire.com'
API_VER = '1.3'

# Retries on connection errors/timeouts
API_ERROR_MAX_RETRIES = 5

logger = logging.getLogger(__name__)

# Each API call may have lots of parameters, so disable warning
# pylint: disable=too-many-arguments


class QueryParams(dict):
    """dict tailored for MediaFire requests.

    * won't store None values
    * boolean values are converted to 'yes'/'no'
    """

    def __init__(self, defaults=None):
        super(QueryParams, self).__init__()
        if defaults is not None:
            for key, value in defaults.items():
                self.__setitem__(key, value)

    def __setitem__(self, key, value):
        """Set dict item, handling booleans"""
        if value is not None:
            if value is True:
                value = 'yes'
            elif value is False:
                value = 'no'
            dict.__setitem__(self, key, value)


class MediaFireError(Exception):
    """Base class for MediaFire-related errors"""
    pass


class MediaFireApiError(MediaFireError):
    """Base class for API errors"""
    def __init__(self, message, code=None):
        """Initialize exception"""
        self.code = code
        self.message = message
        super(MediaFireApiError, self).__init__(message, code)

    def __str__(self):
        """Stringify exception"""
        return "{}: {}".format(self.code, self.message)


class MediaFireConnectionError(MediaFireError):
    """Low level connection errors"""
    pass


class MediaFireApi(object):  # pylint: disable=too-many-public-methods
    """Low-level HTTP API Client"""

    def __init__(self):
        """Initialize MediaFire Client"""

        self.http = requests.Session()
        self.http.mount('https://',
                        HTTPAdapter(max_retries=API_ERROR_MAX_RETRIES))

        self._session = None
        self._action_tokens = {}

    @staticmethod
    def _build_uri(action):
        """Build endpoint URI from action"""
        return '/api/' + API_VER + '/' + action + '.php'

    def _build_query(self, uri, params=None, action_token_type=None):
        """Prepare query string"""

        if params is None:
            params = QueryParams()

        params['response_format'] = 'json'

        session_token = None

        if action_token_type in self._action_tokens:
            # Favor action token
            using_action_token = True
            session_token = self._action_tokens[action_token_type]
        else:
            using_action_token = False
            if self._session:
                session_token = self._session['session_token']

        if session_token:
            params['session_token'] = session_token

        # make order of parameters predictable for testing
        keys = list(params.keys())
        keys.sort()

        query = urlencode([tuple([key, params[key]]) for key in keys])

        if not using_action_token and self._session:
            secret_key_mod = int(self._session['secret_key']) % 256

            signature_base = (str(secret_key_mod) +
                              self._session['time'] +
                              uri + '?' + query).encode('ascii')

            query += '&signature=' + hashlib.md5(signature_base).hexdigest()

        return query

    def request(self, action, params=None, action_token_type=None,
                upload_info=None, headers=None):
        """Perform request to MediaFire API

        action -- "category/name" of method to call
        params -- dict of parameters or query string
        action_token_type -- action token to use: None, "upload", "image"
        upload_info -- in case of upload, dict of "fd" and "filename"
        headers -- additional headers to send (used for upload)

        session_token and signature generation/update is handled automatically
        """

        uri = self._build_uri(action)

        if isinstance(params, six.text_type):
            query = params
        else:
            query = self._build_query(uri, params, action_token_type)

        if headers is None:
            headers = {}

        if upload_info is None:
            # Use request body for query
            data = query
            headers['Content-Type'] = 'application/x-www-form-urlencoded'
        else:
            # Use query string for query since payload is file
            uri += '?' + query

            if "filename" in upload_info:
                data = MultipartEncoder(
                    fields={'file': (
                        upload_info["filename"],
                        upload_info["fd"],
                        'application/octet-stream'
                    )}
                )
                headers["Content-Type"] = data.content_type
            else:
                data = upload_info["fd"]
                headers["Content-Type"] = 'application/octet-stream'

        logger.debug("uri=%s query=%s",
                     uri, query if not upload_info else None)

        try:
            response = self.http.post(API_BASE + uri, data=data,
                                      headers=headers, stream=True)
        except RequestException as ex:
            logger.exception("HTTP request failed")
            raise MediaFireConnectionError(
                "RequestException: {}".format(ex))

        return self._process_response(response)

    def _process_response(self, response):
        """Parse response"""

        forward_raw = False
        content_type = response.headers['Content-Type']
        if content_type != 'application/json':
            logger.debug("headers: %s", response.headers)
            # API BUG: text/xml content-type with json payload
            # http://forum.mediafiredev.com/showthread.php?136
            if content_type == 'text/xml':
                # we never request xml
                if response.text.lstrip().startswith('{'):
                    logger.debug("API BUG: text/xml content-type "
                                 "with JSON payload")
                else:
                    forward_raw = True
            else:
                # _process_response can't deal with non-json,
                # return response as is
                forward_raw = True

        if forward_raw:
            response.raise_for_status()
            return response

        logger.debug("response: %s", response.text)

        # if we are here, then most likely have json
        try:
            response_node = response.json()['response']
        except ValueError:
            # promised JSON but failed
            raise MediaFireApiError("JSON decode failure")

        if response_node.get('new_key', 'no') == 'yes':
            self._regenerate_secret_key()

        # check for errors
        if response_node['result'] != 'Success':
            raise MediaFireApiError(response_node['message'],
                                    response_node['error'])

        return response_node

    def _regenerate_secret_key(self):
        """Regenerate secret key

        http://www.mediafire.com/developers/core_api/1.3/getting_started/#call_signature
        """
        # Don't regenerate the key if we have none
        if self._session and 'secret_key' in self._session:
            self._session['secret_key'] = (
                int(self._session['secret_key']) * 16807) % 2147483647

    @property
    def session(self):
        """Returns current session information"""
        return self._session

    @session.setter
    def session(self, value):
        """Set session token

        value -- dict returned by user/get_session_token"""

        # unset session token
        if value is None:
            self._session = None
            return

        if not isinstance(value, dict):
            raise ValueError("session info is required")

        session_parsed = {}

        for key in ["session_token", "time", "secret_key"]:
            if key not in value:
                raise ValueError("Missing parameter: {}".format(key))
            session_parsed[key] = value[key]

        for key in ["ekey", "pkey"]:
            # nice to have, but not mandatory
            if key in value:
                session_parsed[key] = value[key]

        self._session = session_parsed

    @session.deleter
    def session(self):
        """Unset session"""
        self._session = None

    def set_action_token(self, type_=None, action_token=None):
        """Set action tokens

        type_ -- either "upload" or "image"
        action_token -- string obtained from user/get_action_token,
                        set None to remove the token
        """
        if action_token is None:
            del self._action_tokens[type_]
        else:
            self._action_tokens[type_] = action_token

    def user_fetch_tos(self):
        """user/fetch_tos

        http://www.mediafire.com/developers/core_api/1.3/user/#fetch_tos
        """

        return self.request("user/fetch_tos")

    def user_accept_tos(self, acceptance_token):
        """user/accept_tos

        http://www.mediafire.com/developers/core_api/1.3/user/#user_top
        """

        return self.request("user/accept_tos", QueryParams({
            "acceptance_token": acceptance_token
        }))

    def user_get_session_token(self, app_id=None, email=None, password=None,
                               ekey=None, fb_access_token=None,
                               tw_oauth_token=None,
                               tw_oauth_token_secret=None, api_key=None):
        """user/get_session_token

        http://www.mediafire.com/developers/core_api/1.3/user/#get_session_token
        """

        if app_id is None:
            raise ValueError("app_id must be defined")

        params = QueryParams({
            'application_id': str(app_id),
            'token_version': 2,
            'response_format': 'json'
        })

        if fb_access_token:
            params['fb_access_token'] = fb_access_token
            signature_keys = ['fb_access_token']
        elif tw_oauth_token and tw_oauth_token_secret:
            params['tw_oauth_token'] = tw_oauth_token
            params['tw_oauth_token_secret'] = tw_oauth_token_secret
            signature_keys = ['tw_oauth_token',
                              'tw_oauth_token_secret']
        elif (email or ekey) and password:
            signature_keys = []
            if email:
                signature_keys.append('email')
                params['email'] = email

            if ekey:
                signature_keys.append('ekey')
                params['ekey'] = ekey

            params['password'] = password
            signature_keys.append('password')
        else:
            raise ValueError("Credentials not provided")

        signature_keys.append('application_id')

        signature = hashlib.sha1()
        for key in signature_keys:
            signature.update(str(params[key]).encode('ascii'))

        # Note: If the app uses a callback URL to provide its API key,
        # or if it does not have the "Require Secret Key" option checked,
        # then the API key may be omitted from the signature
        if api_key:
            signature.update(api_key.encode('ascii'))

        query = urlencode(params)
        query += '&signature=' + signature.hexdigest()

        return self.request('user/get_session_token', params=query)

    def user_renew_session_token(self):
        """user/renew_session_token:

        http://www.mediafire.com/developers/core_api/1.3/user/#renew_session_token
        """
        return self.request('user/renew_session_token')

    def user_get_action_token(self, type_=None, lifespan=None):
        """user/get_action_token

        http://www.mediafire.com/developers/core_api/1.3/user/#get_action_token
        """
        return self.request('user/get_action_token', QueryParams({
            'type': type_,
            'lifespan': lifespan
        }))

    def user_destroy_action_token(self, action_token=None):
        """user/destroy_action_token

        http://www.mediafire.com/developers/core_api/1.3/user/#destroy_action_token
        """
        return self.request('user/destroy_action_token', QueryParams({
            'action_token': action_token
        }))

    def user_get_avatar(self):
        """user/get_avatar

        http://www.mediafire.com/developers/core_api/1.3/user/#get_avatar
        """
        return self.request("user/get_avatar")

    def user_get_info(self):
        """user/get_info

        http://www.mediafire.com/developers/core_api/1.3/user/#get_info
        """
        return self.request("user/get_info")

    def user_get_limits(self):
        """user/get_limits

        http://www.mediafire.com/developers/core_api/1.3/user/#get_limits
        """
        return self.request("user/get_limits")

    def user_get_settings(self):
        """user/get_settings

        http://www.mediafire.com/developers/core_api/1.3/user/#get_settings
        """
        return self.request("user/get_settings")

    def user_set_avatar(self, action=None, quick_key=None, url=None):
        """user/set_avatar

        http://www.mediafire.com/developers/core_api/1.3/user/#set_avatar
        """
        return self.request("user/set_avatar", QueryParams({
            "action": action,
            "quick_key": quick_key,
            "url": url
        }))

    def user_update(self, display_name=None, first_name=None, last_name=None,
                    email=None, password=None, current_password=None,
                    birth_date=None, gender=None, website=None, subdomain=None,
                    location=None, newsletter=None, primary_usage=None,
                    timezone=None):
        """
        user/update

        http://www.mediafire.com/developers/core_api/1.3/user/#update
        """
        return self.request("user/update", QueryParams({
            "display_name": display_name,
            "first_name": first_name,
            "last_name": last_name,
            "email": email,
            "password": password,
            "current_password": current_password,
            "birth_date": birth_date,
            "gender": gender,
            "website": website,
            "subdomain": subdomain,
            "location": location,
            "newsletter": newsletter,
            "primary_usage": primary_usage,
            "timezone": timezone
        }))

    def folder_get_info(self, folder_key=None, device_id=None, details=None):
        """folder/get_info

        http://www.mediafire.com/developers/core_api/1.3/folder/#get_info
        """
        return self.request('folder/get_info', QueryParams({
            'folder_key': folder_key,
            'device_id': device_id,
            'details': details
        }))

    def folder_get_content(self, folder_key=None, content_type=None,
                           filter_=None, device_id=None, order_by=None,
                           order_direction=None, chunk=None, details=None,
                           chunk_size=None):
        """folder/get_content

        http://www.mediafire.com/developers/core_api/1.3/folder/#get_content
        """
        return self.request('folder/get_content', QueryParams({
            'folder_key': folder_key,
            'content_type': content_type,
            'filter': filter_,
            'device_id': device_id,
            'order_by': order_by,
            'order_direction': order_direction,
            'chunk': chunk,
            'details': details,
            'chunk_size': chunk_size
        }))

    def folder_update(self, folder_key, foldername=None, description=None,
                      privacy=None, privacy_recursive=None, mtime=None):
        """folder/update

        http://www.mediafire.com/developers/core_api/1.3/folder/#update
        """
        return self.request('folder/update', QueryParams({
            'folder_key': folder_key,
            'foldername': foldername,
            'description': description,
            'privacy': privacy,
            'privacy_recursive': privacy_recursive,
            'mtime': mtime
        }))

    def folder_create(self, foldername=None, parent_key=None,
                      action_on_duplicate=None, mtime=None):
        """folder/create

        http://www.mediafire.com/developers/core_api/1.3/folder/#create
        """
        return self.request('folder/create', QueryParams({
            'foldername': foldername,
            'parent_key': parent_key,
            'action_on_duplicate': action_on_duplicate,
            'mtime': mtime
        }))

    def folder_delete(self, folder_key):
        """folder/delete

        http://www.mediafire.com/developers/core_api/1.3/folder/#delete
        """
        return self.request('folder/delete', QueryParams({
            'folder_key': folder_key
        }))

    def folder_purge(self, folder_key):
        """folder/purge

        http://www.mediafire.com/developers/core_api/1.3/folder/#purge
        """
        return self.request('folder/purge', QueryParams({
            'folder_key': folder_key
        }))

    def folder_move(self, folder_key_src, folder_key_dst=None):
        """folder/move

        http://www.mediafire.com/developers/core_api/1.3/folder/#move
        """
        return self.request('folder/move', QueryParams({
            'folder_key_src': folder_key_src,
            'folder_key_dst': folder_key_dst
        }))

    def upload_check(self, filename=None, folder_key=None, filedrop_key=None,
                     size=None, hash_=None, path=None, resumable=None):
        """upload/check

        http://www.mediafire.com/developers/core_api/1.3/upload/#check
        """
        return self.request('upload/check', QueryParams({
            'filename': filename,
            'folder_key': folder_key,
            'filedrop_key': filedrop_key,
            'size': size,
            'hash': hash_,
            'path': path,
            'resumable': resumable
        }))

    def upload_simple(self, fd, filename, folder_key=None, path=None,
                      filedrop_key=None, action_on_duplicate=None,
                      mtime=None, file_size=None, file_hash=None):
        """upload/simple

        http://www.mediafire.com/developers/core_api/1.3/upload/#simple
        """
        action = 'upload/simple'

        params = QueryParams({
            'folder_key': folder_key,
            'path': path,
            'filedrop_key': filedrop_key,
            'action_on_duplicate': action_on_duplicate,
            'mtime': mtime
        })

        headers = QueryParams({
            'X-Filesize': file_size,
            'X-Filehash': file_hash,
            'X-Filename': filename.encode('utf-8')
        })

        upload_info = {
            "fd": fd,
        }

        return self.request(action, params, action_token_type="upload",
                            upload_info=upload_info, headers=headers)

    # pylint: disable=too-many-locals
    # The API requires us to provide all of that
    def upload_resumable(self, fd, filesize, filehash, unit_hash, unit_id,
                         unit_size, quick_key=None, action_on_duplicate=None,
                         mtime=None, version_control=None, folder_key=None,
                         filedrop_key=None, path=None, previous_hash=None):
        """upload/resumable

        http://www.mediafire.com/developers/core_api/1.3/upload/#resumable
        """
        action = 'upload/resumable'

        headers = {
            'x-filesize': filesize,
            'x-filehash': filehash,
            'x-unit-hash': unit_hash,
            'x-unit-id': unit_id,
            'x-unit-size': unit_size
        }

        params = QueryParams({
            'quick_key': quick_key,
            'action_on_duplicate': action_on_duplicate,
            'mtime': mtime,
            'version_control': version_control,
            'folder_key': folder_key,
            'filedrop_key': filedrop_key,
            'path': path,
            'previous_hash': previous_hash
        })

        upload_info = {
            "fd": fd,
            "filename": "chunk"
        }

        return self.request(action, params, action_token_type="upload",
                            upload_info=upload_info, headers=headers)
    # pylint: enable=too-many-locals

    def upload_instant(self, filename, size, hash_, quick_key=None,
                       folder_key=None, filedrop_key=None, path=None,
                       action_on_duplicate=None, mtime=None,
                       version_control=None, previous_hash=None):
        """upload/instant

        http://www.mediafire.com/developers/core_api/1.3/upload/#instant
        """
        return self.request('upload/instant', QueryParams({
            'filename': filename,
            'size': size,
            'hash': hash_,
            'quick_key': quick_key,
            'folder_key': folder_key,
            'filedrop_key': filedrop_key,
            'path': path,
            'action_on_duplicate': action_on_duplicate,
            'mtime': mtime,
            'version_control': version_control,
            'previous_hash': previous_hash
        }))

    def upload_poll(self, key):
        """upload/poll

        http://www.mediafire.com/developers/core_api/1.3/upload/#poll_upload
        """
        return self.request('upload/poll_upload', QueryParams({
            'key': key
        }))

    def file_get_info(self, quick_key=None):
        """file/get_info

        http://www.mediafire.com/developers/core_api/1.3/file/#get_info
        """
        return self.request('file/get_info', QueryParams({
            'quick_key': quick_key
        }))

    def file_get_links(self, quick_key, link_type=None):
        """file/get_links

        http://www.mediafire.com/developers/core_api/1.3/file/#get_links
        """
        return self.request('file/get_links', QueryParams({
            'quick_key': quick_key,
            'link_type': link_type,
        }))

    def file_update(self, quick_key, filename=None, description=None,
                    mtime=None, privacy=None):
        """file/update

        http://www.mediafire.com/developers/core_api/1.3/file/#update
        """
        return self.request('file/update', QueryParams({
            'quick_key': quick_key,
            'filename': filename,
            'description': description,
            'mtime': mtime,
            'privacy': privacy
        }))

    def file_update_file(self, quick_key, file_extension=None, filename=None,
                         description=None, mtime=None, privacy=None,
                         timezone=None):
        """file/update_file

        http://www.mediafire.com/developers/core_api/1.3/file/#update_file
        """
        return self.request('file/update', QueryParams({
            'quick_key': quick_key,
            'file_extension': file_extension,
            'filename': filename,
            'description': description,
            'mtime': mtime,
            'privacy': privacy,
            'timezone': timezone
        }))

    def file_delete(self, quick_key):
        """file/delete

        http://www.mediafire.com/developers/core_api/1.3/file/#delete
        """
        return self.request('file/delete', QueryParams({
            'quick_key': quick_key
        }))

    def file_move(self, quick_key, folder_key=None):
        """file/move

        http://www.mediafire.com/developers/core_api/1.3/file/#move
        """
        return self.request('file/move', QueryParams({
            'quick_key': quick_key,
            'folder_key': folder_key
        }))

    def file_purge(self, quick_key):
        """file/purge

        http://www.mediafire.com/developers/core_api/1.3/file/#purge
        """
        return self.request('file/purge', QueryParams({
            'quick_key': quick_key
        }))

    def file_zip(self, keys, confirm_download=None, meta_only=None):
        """file/zip

        http://www.mediafire.com/developers/core_api/1.3/file/#zip
        """
        return self.request('file/zip', QueryParams({
            'keys': keys,
            'confirm_download': confirm_download,
            'meta_only': meta_only
        }))

    def system_get_info(self):
        """system/get_info

        http://www.mediafire.com/developers/core_api/1.3/system/#get_info
        """
        return self.request('system/get_info')

    def system_get_status(self):
        """system/get_status

        http://www.mediafire.com/developers/core_api/1.3/system/#get_status
        """
        return self.request('system/get_status')



"""MediaFire Python Open SDK"""

__all__ = ["MediaFireApi",
           "MediaFireApiError",
           "MediaFireUploader",
           "UploadSession"]

from mediafire.api import (MediaFireApi, MediaFireApiError)
from mediafire.uploader import (MediaFireUploader, UploadSession)
# The client, media has not yet graduated
# from mediafire.client import (MediaFireClient, MediaFireError)
# from mediafire.media import ConversionServerClient
